#!/usr/local/bin/python

# Example of shared memory IPC: Producer/Consumer - semaphore implementation

# The parent process allocates semaphores and a shared memory buffer, then
# forks NB_PROD producer and NB_CONS consumer processes. Each forked process
# enters a rendez-vous barrier before competing. Competition starts when
# the parent process enters the rendez-vous barrier. After TIME seconds,
# the parent process kills its children (how cruel!) and frees the allocated
# shared resources.

import whrandom, shm, string, struct, sys

# from the obsolete 'rand' module

def srand(seed):
    whrandom.seed(seed%256, seed/256%256, seed/65536%256)

def rand():
    return int(whrandom.random() * 32768.0) % 32768

# --------------------------------------------------------------------------

KEY_MEMORY = 100	# shared memory segment key

KEY_FULL = 100		# full semaphore key
KEY_EMPTY = 101		# empty semaphore key
KEY_MUTEX = 102		# mutex semaphore key
KEY_BARRIER = 103	# barrier semaphore key

ID = None		# prod/cons ID
TIME = 1		# working time in seconds
NB_PROD = 5		# number of producers
NB_CONS = 5		# number of consumers
BUFF_SIZE = 5		# buffer capacity (nb of items)
ITEM_SIZE = struct.calcsize('i')

def semaphores():
    sems = []
    for semkey, initval in [(KEY_FULL, 0),
			    (KEY_EMPTY, BUFF_SIZE),
			    (KEY_MUTEX, 1),
			    (KEY_BARRIER, NB_PROD + NB_CONS + 1)]:
	try:
	    if shm.semaphore_haskey(semkey):
		s = shm.semaphore(shm.getsemid(semkey))
		s.setval(initval)
	    else:
		s = shm.create_semaphore(semkey, initval)
	except:
	    print "Can't access/allocate semaphore with key %s" % semkey
	    sys.exit(-1)
	sems.append(s)
	s.setundo(1)
    return tuple(sems)

def shared_memory():
    try:
	if shm.memory_haskey(KEY_MEMORY):
	    m = shm.memory(shm.getshmid(KEY_MEMORY))
	else:
	    m = shm.create_memory(KEY_MEMORY, (BUFF_SIZE + 1) * ITEM_SIZE)
    except:
	print "Can't access/allocate shared memory with key %s" % semkey
	sys.exit(-1)
    m.attach()
    return m

def rendez_vous(barrier):
    print "%s: Entering rendez-vous barrier" % ID
    barrier.P()
    barrier.Z()

def get_item(buffer):
    nbitems, = struct.unpack('i', buffer.read(ITEM_SIZE))
    item, = struct.unpack('i', buffer.read(ITEM_SIZE, nbitems * ITEM_SIZE))
    nbitems = nbitems - 1
    buffer.write(struct.pack('i', nbitems))
    print "%s: get %5s, items in buffer: %s" % (ID, `item`, nbitems)
    return item

def put_item(buffer, item):
    nbitems, = struct.unpack('i', buffer.read(ITEM_SIZE))
    for i in range(nbitems, 0, -1):	# Move items to preserve FIFO order
	buffer.write(buffer.read(ITEM_SIZE, i * ITEM_SIZE), (i+1) * ITEM_SIZE)
    buffer.write(struct.pack('i', item), 1 * ITEM_SIZE)
    nbitems = nbitems + 1
    buffer.write(struct.pack('i', nbitems))
    print "%s: put %5s, items in buffer: %s" % (ID, `item`, nbitems)
    
def produce(full, empty, mutex, buffer):
    item = rand()	# produce an item
    empty.P()
    mutex.P()
    put_item(buffer, item)
    mutex.V()
    full.V()

def consume(full, empty, mutex, buffer):
    full.P()
    mutex.P()
    item = get_item(buffer)
    mutex.V()
    empty.V()
    None	# consume the item here

def cleanup():
    for semkey in [KEY_FULL, KEY_EMPTY, KEY_MUTEX, KEY_BARRIER]:
	try:
	    shm.remove_semaphore(shm.getsemid(semkey))
	except:
	    print "Can't remove semaphore with key %s" % semkey
    try:
	shm.remove_memory(shm.getshmid(KEY_MEMORY))
    except:
	print "Can't remove shared memory segment with key %s" % KEY_MEMORY

def usage():
    print "Usage: python prodcons.py [NB_PROD=5 [NB_CONS=5 [BUFF_SIZE=5 [TIME=1]]]]"
    sys.exit(1)

def main(argv):
    if len(argv) > 5:
	usage()
    if len(argv) > 4:
	global TIME
	try:
	    TIME = abs(string.atof(argv[4]))
	except:
	    usage()
    if len(argv) > 3:
	global BUFF_SIZE
	try:
	    BUFF_SIZE = abs(string.atoi(argv[3]))
	except:
	    usage()
    if len(argv) > 2:
	global NB_CONS
	try:
	    NB_CONS = abs(string.atoi(argv[2]))
	except:
	    usage()
    if len(argv) > 1:
	global NB_PROD
	try:
	    NB_PROD = abs(string.atoi(argv[1]))
	except:
	    usage()

    full, empty, mutex, barrier = semaphores()
    buffer = shared_memory()
    buffer.write(struct.pack('i', 0))	# first entry (nb of items in buffer)

    import os, time
    children=[]
    for producer in range(NB_PROD):
	time.sleep(0.2)
	child = os.fork()
	if child == 0:
	    global ID
	    ID = "Producer %s" % (producer + 1)
	    srand(hash(ID))
	    rendez_vous(barrier)
	    while 1:
		produce(full, empty, mutex, buffer)
	else:
	    children.append(child)
    for consumer in range(NB_CONS):
	time.sleep(0.2)
	child = os.fork()
	if child == 0:
	    global ID
	    ID = "Consumer %s" % (consumer + 1)
	    rendez_vous(barrier)
	    while 1:
		consume(full, empty, mutex, buffer)
	else:
	    children.append(child)
    print "START!"
    time.sleep(0.5)
    barrier.P()
    time.sleep(TIME)
    mutex.P()
    for child in children:
	os.kill(child, 9)
    mutex.V()
    cleanup()
    print "End."

if __name__ == "__main__":
    main(sys.argv)
